"""Asynchronous Python client providing RDW vehicle information."""
from pathlib import Path


def load_fixture(filename: str) -> str:
    """Load a fixture."""
    path = Path(__package__) / "fixtures" / filename
    return path.read_text(encoding="utf-8")



"""Tests for the vehicle Library."""
# pylint: disable=protected-access
import asyncio

import aiohttp
import pytest
from aresponses import Response, ResponsesMockServer

from vehicle import RDW
from vehicle.const import Dataset
from vehicle.exceptions import RDWConnectionError, RDWError


async def test_json_request(aresponses: ResponsesMockServer) -> None:
    """Test JSON response is handled correctly."""
    aresponses.add(
        "opendata.rdw.nl",
        "/resource/m9d7-ebf2.json",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"status": "ok"}',
        ),
    )
    async with aiohttp.ClientSession() as session:
        rdw = RDW(session=session)
        response = await rdw._request(Dataset.PLATED_VEHICLES)
        assert response == '{"status": "ok"}'
        await rdw.close()


async def test_internal_session(aresponses: ResponsesMockServer) -> None:
    """Test JSON response is handled correctly."""
    aresponses.add(
        "opendata.rdw.nl",
        "/resource/m9d7-ebf2.json",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text='{"status": "ok"}',
        ),
    )
    async with RDW() as rdw:
        response = await rdw._request(Dataset.PLATED_VEHICLES)
        assert response == '{"status": "ok"}'


async def test_timeout(aresponses: ResponsesMockServer) -> None:
    """Test request timeout."""

    # Faking a timeout by sleeping
    async def response_handler(_: aiohttp.ClientResponse) -> Response:
        """Response handler for this test."""
        await asyncio.sleep(2)
        return aresponses.Response(body="Goodmorning!")

    aresponses.add(
        "opendata.rdw.nl",
        "/resource/m9d7-ebf2.json",
        "GET",
        response_handler,
    )

    async with aiohttp.ClientSession() as session:
        rdw = RDW(session=session, request_timeout=1)
        with pytest.raises(RDWConnectionError):
            assert await rdw._request(Dataset.PLATED_VEHICLES)


async def test_http_error400(aresponses: ResponsesMockServer) -> None:
    """Test HTTP 404 response handling."""
    aresponses.add(
        "opendata.rdw.nl",
        "/resource/m9d7-ebf2.json",
        "GET",
        aresponses.Response(text="OMG PUPPIES!", status=404),
    )

    async with aiohttp.ClientSession() as session:
        rdw = RDW(session=session)
        with pytest.raises(RDWError):
            assert await rdw._request(Dataset.PLATED_VEHICLES)


async def test_unexpected_response(aresponses: ResponsesMockServer) -> None:
    """Test unexpected response handling."""
    aresponses.add(
        "opendata.rdw.nl",
        "/resource/m9d7-ebf2.json",
        "GET",
        aresponses.Response(text="OMG PUPPIES!", status=200),
    )

    async with aiohttp.ClientSession() as session:
        rdw = RDW(session=session)
        with pytest.raises(RDWError):
            assert await rdw._request(Dataset.PLATED_VEHICLES)


async def test_license_plate_normalization() -> None:
    """Test normalization of license plates."""
    assert RDW.normalize_license_plate("AB-12-34") == "AB1234"
    assert RDW.normalize_license_plate("AB1234") == "AB1234"
    assert RDW.normalize_license_plate("AB1-23-4") == "AB1234"
    assert RDW.normalize_license_plate(" Ab1-23-4 ") == "AB1234"



"""Asynchronous Python client providing RDW vehicle information."""

from enum import Enum


class Dataset(str, Enum):
    """Enum holding the RDW dataset identifiers for the Socrata API."""

    PLATED_VEHICLES = "m9d7-ebf2"


class VehicleType(str, Enum):
    """Enum holding the RDW vehicle types."""

    TAILER = "Aanhangwagen"
    PASSENGER_CAR = "Personenauto"
    COMPANY_CAR = "Bedrijfsauto"
    BUS = "Bus"
    MOTORCYCLE = "Motorfiets"
    THREE_WHEELED_MOTOR_VEHICLE = "Driewielig motorrijtuig"
    MOPED = "Bromfiets"
    CENTER_AXLE_TRAILER = "Middenasaanhangwagen"
    AGRICULTURAL_OR_FORESTRY_TRACTOR = "Land- of bosbouwtrekker"
    AGRICULTURAL_OR_FORESTRY_TRACTOR_TRAILER = "Land- of bosb aanhw of getr uitr stuk"


class VehicleInterior(str, Enum):
    """Enum holding the RDW vehicle interior types."""

    CAMPER = "kampeerwagen"
    CARAVAN = "caravan"
    CLOSED_INTERIOR = "gesloten opbouw"
    CONVERTIBLE = "cabriolet"
    COUPE = "coupe"
    FIRETRUCK = "brandweerwagen"
    HATCHBACK = "hatchback"
    LIVESTOCK_TRUCK = "veewagen"
    MPV = "MPV"
    OPEN_LOADING_FLOOR = "open laadvloer"
    OPEN_VEHICLE = "open wagen"
    SEDAN = "sedan"
    STATION_WAGON = "stationwagen"
    PICK_UP_TRUCK = "pick-up truck"


class VehicleOdometerJudgement(str, Enum):
    """Enum holding the RDW vehicle odometer judgement types."""

    NO_JUDGEMENT = "Geen oordeel"
    LOGICAL = "Logisch"
    ILLOGICAL = "Onlogisch"
    UNKNOWN = "UNKNOWN"



"""Asynchronous Python client providing RDW vehicle information."""
from __future__ import annotations

import asyncio
import socket
from dataclasses import dataclass
from importlib import metadata
from typing import Any, Self

import orjson
from aiohttp.client import ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import METH_GET
from yarl import URL

from .const import Dataset
from .exceptions import RDWConnectionError, RDWError, RDWUnknownLicensePlateError
from .models import Vehicle


@dataclass
class RDW:
    """Main class for handling data fetching from RDW."""

    session: ClientSession | None = None
    license_plate: str | None = None
    request_timeout: int = 10
    _close_session: bool = False

    @staticmethod
    def normalize_license_plate(license_plate: str) -> str:
        """Normalize license plate.

        Args:
        ----
            license_plate: License plate.

        Returns:
        -------
            Normalized license plate.
        """
        return license_plate.upper().replace("-", "").replace(" ", "").strip()

    async def _request(
        self,
        dataset: Dataset,
        *,
        data: dict[str, Any] | None = None,
    ) -> str:
        """Handle a request to a RDW open data (Socrata).

        A generic method for sending/handling HTTP requests done against
        the public RDW data.

        Args:
        ----
            dataset: Identifier for the Socrata dataset to query.
            data: Dictionary of data to send to the Socrata API.

        Returns:
        -------
            A Python dictionary (JSON decoded) with the response from
            the API.

        Raises:
        ------
            RDWConnectionError: An error occurred while communicating with
                the Socrata API.
            RDWError: Received an unexpected response from the Socrata API.
        """
        version = metadata.version(__package__)
        url = URL("https://opendata.rdw.nl/resource/").join(
            URL(f"{dataset.value}.json"),
        )

        headers = {
            "User-Agent": f"PythonVehicle/{version}",
            "Accept": "application/json, text/plain, */*",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    METH_GET,
                    url.with_query(data),
                    headers=headers,
                )
                response.raise_for_status()
        except asyncio.TimeoutError as exception:
            msg = "Timeout occurred while connecting to the Socrata API"
            raise RDWConnectionError(msg) from exception
        except (
            ClientError,
            ClientResponseError,
            socket.gaierror,
        ) as exception:
            msg = "Error occurred while communicating with Socrata API"
            raise RDWConnectionError(msg) from exception

        content_type = response.headers.get("Content-Type", "")
        text = await response.text()
        if "application/json" not in content_type:
            msg = "Unexpected response from the Socrata API"
            raise RDWError(
                msg,
                {"Content-Type": content_type, "response": text},
            )

        return text

    async def vehicle(self, license_plate: str | None = None) -> Vehicle:
        """Get devices information about a Vehicle.

        Args:
        ----
            license_plate: License plate of the vehicle, if not provided
                the license plate of the object will be used.

        Returns:
        -------
            A Vehicle object, with information about the vehicle.

        Raises:
        ------
            RDWError: No license plate provided.
            RDWUnknownLicensePlateError: License plate not found in RDW Socrata DB.
        """
        license_plate = license_plate or self.license_plate
        if license_plate is None:
            msg = "No license plate provided"
            raise RDWError(msg)

        data = await self._request(
            Dataset.PLATED_VEHICLES,
            data={"kenteken": self.normalize_license_plate(license_plate)},
        )
        # pylint: disable=no-member
        vehicles = orjson.loads(data)
        if not vehicles:
            msg = f"License plate {license_plate} not found in RDW Socrata database"
            raise RDWUnknownLicensePlateError(msg)

        return Vehicle.from_dict(vehicles[0])

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The RDW object.
        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.
        """
        await self.close()



"""Asynchronous Python client providing RDW vehicle information."""


class RDWError(Exception):
    """Generic exception."""


class RDWUnknownLicensePlateError(RDWError):
    """RDW connection exception."""


class RDWConnectionError(RDWError):
    """RDW connection exception."""



# pylint: disable=W0621
"""Asynchronous Python client providing RDW vehicle information."""

import asyncio

from vehicle import RDW, Vehicle


async def main() -> None:
    """Show example of fetching RDW vehicle info from Socrata API."""
    async with RDW() as rdw:
        vehicle: Vehicle = await rdw.vehicle(license_plate="AR6458")
        print(vehicle)


if __name__ == "__main__":
    asyncio.run(main())



"""Tests for the vehicle Library."""

import aiohttp
import pytest
from aresponses import ResponsesMockServer
from syrupy import SnapshotAssertion

from vehicle import (
    RDW,
    RDWError,
    RDWUnknownLicensePlateError,
    Vehicle,
)

from . import load_fixture


@pytest.mark.parametrize(
    "license_plate",
    [
        "11ZKZ3",
        "0001TJ",
        "VXJ99N",
    ],
)
async def test_vehicle_data(
    aresponses: ResponsesMockServer,
    license_plate: str,
    snapshot: SnapshotAssertion,
) -> None:
    """Test getting Vehicle information."""
    aresponses.add(
        "opendata.rdw.nl",
        "/resource/m9d7-ebf2.json",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture(f"{license_plate}.json"),
        ),
    )
    async with aiohttp.ClientSession() as session:
        rdw = RDW(session=session)
        vehicle: Vehicle = await rdw.vehicle(license_plate)
        assert vehicle == snapshot
        assert vehicle.to_json() == snapshot


async def test_no_vehicle(aresponses: ResponsesMockServer) -> None:
    """Test getting non-existing Vehicle."""
    aresponses.add(
        "opendata.rdw.nl",
        "/resource/m9d7-ebf2.json",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("no_vehicles.json"),
        ),
    )
    async with aiohttp.ClientSession() as session:
        rdw = RDW(session=session)
        with pytest.raises(RDWUnknownLicensePlateError):
            await rdw.vehicle("00-00-00")


async def test_no_license_plate_provided() -> None:
    """Test getting Vehicle without license plate."""
    rdw = RDW()
    with pytest.raises(RDWError):
        await rdw.vehicle()



"""Asynchronous Python client providing RDW vehicle information."""
from .const import VehicleInterior, VehicleOdometerJudgement, VehicleType
from .exceptions import RDWConnectionError, RDWError, RDWUnknownLicensePlateError
from .models import Vehicle
from .rdw import RDW

__all__ = [
    "RDW",
    "RDWConnectionError",
    "RDWError",
    "RDWUnknownLicensePlateError",
    "Vehicle",
    "VehicleInterior",
    "VehicleOdometerJudgement",
    "VehicleType",
]



"""Asynchronous Python client providing RDW vehicle information."""
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date, datetime, timezone
from typing import Any

from mashumaro import field_options
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import SerializationStrategy

from .const import (
    VehicleInterior,
    VehicleOdometerJudgement,
    VehicleType,
)


class StringIsBoolean(SerializationStrategy):
    """Boolean serialization strategy for Dutch textual strings."""

    def serialize(self, value: bool) -> str:  # noqa: FBT001
        """Serialize a boolean to an Dutch string."""
        return "Ja" if value else "Nee"

    def deserialize(self, value: str) -> bool:
        """Deserialize an Dutch string to a boolean."""
        return value == "Ja"


class DateStrategy(SerializationStrategy):
    """String serialization strategy to handle the date format."""

    def serialize(self, value: date) -> str:
        """Serialize date to their specific format."""
        return datetime.strftime(value, "%Y%m%d")

    def deserialize(self, value: str) -> date:
        """Deserialize their date format to a date."""
        return datetime.strptime(value, "%Y%m%d").replace(tzinfo=timezone.utc).date()


@dataclass
# pylint: disable-next=too-many-instance-attributes
class Vehicle(DataClassORJSONMixin):
    """Object holding vehicle information.

    Attributes
    ----------
        apk_expiration: Expiry date of the APK.
        ascription_date: Date of naming registration of the vehicle.
        ascription_possible: Whether the vehicle is nameable or not.
        brand: Brand of the vehicle.
        energy_label: Energy label of the vehicle.
        engine_capacity: Engine capacity of the vehicle in CC
        exported: Whether the vehicle is exported or not.
        first_admission: First date of registration.
        interior: Interior of the vehicle.
        last_odometer_registration_year: Last year odometer was registered.
        liability_insured: Whether the vehicle is insured or not.
        license_plate: Normalized license plate of the vehicle.
        list_price: List price of the vehicle.
        mass_driveable: Mass of the vehicle when driveable in KG.
        mass_empty: Empty mass of the vehicle in KG.
        model: Model of the vehicle.
        number_of_cylinders: Number of cylinders of the vehicle.
        number_of_doors: Number of doors of the vehicle.
        number_of_seats: Number of seats of the vehicle.
        number_of_wheelchair_seats: Number of wheelchair seats of the vehicle.
        number_of_wheels: Number of wheels of the vehicle.
        odometer_judgement: Odometer judgement of the vehicle.
        pending_recall: Whether the vehicle has a pending recall or not.
        taxi: Whether the vehicle is a taxi or not.
        vehicle_type: Type of the vehicle.
    """

    # pylint: disable-next=too-few-public-methods
    class Config(BaseConfig):
        """Mashumaro configuration."""

        serialization_strategy = {bool: StringIsBoolean(), date: DateStrategy()}  # noqa: RUF012
        serialize_by_alias = True

    brand: str = field(metadata=field_options(alias="merk"))
    license_plate: str = field(metadata=field_options(alias="kenteken"))
    model: str = field(metadata=field_options(alias="handelsbenaming"))
    apk_expiration: date | None = field(
        default=None, metadata=field_options(alias="vervaldatum_apk")
    )
    ascription_date: date | None = field(
        default=None, metadata=field_options(alias="datum_tenaamstelling")
    )
    ascription_possible: bool | None = field(
        default=None, metadata=field_options(alias="tenaamstellen_mogelijk")
    )
    energy_label: str | None = field(
        default=None, metadata=field_options(alias="zuinigheidslabel")
    )
    engine_capacity: int | None = field(
        default=None, metadata=field_options(alias="cilinderinhoud")
    )
    exported: bool | None = field(
        default=None, metadata=field_options(alias="export_indicator")
    )
    interior: VehicleInterior | None = field(
        default=None, metadata=field_options(alias="inrichting")
    )
    last_odometer_registration_year: int | None = field(
        default=None,
        metadata=field_options(alias="jaar_laatste_registratie_tellerstand"),
    )
    liability_insured: bool | None = field(
        default=None, metadata=field_options(alias="wam_verzekerd")
    )
    list_price: int | None = field(
        default=None, metadata=field_options(alias="catalogusprijs")
    )
    first_admission: date | None = field(
        default=None, metadata=field_options(alias="datum_eerste_toelating")
    )
    mass_empty: int | None = field(
        default=None, metadata=field_options(alias="massa_ledig_voertuig")
    )
    mass_driveable: int | None = field(
        default=None, metadata=field_options(alias="massa_rijklaar")
    )
    number_of_cylinders: int | None = field(
        default=None, metadata=field_options(alias="aantal_cilinders")
    )
    number_of_doors: int | None = field(
        default=None, metadata=field_options(alias="aantal_deuren")
    )
    number_of_seats: int | None = field(
        default=None, metadata=field_options(alias="aantal_zitplaatsen")
    )
    number_of_wheelchair_seats: int | None = field(
        default=None,
        metadata=field_options(alias="aantal_rolstoelplaatsen"),
    )
    number_of_wheels: int | None = field(
        default=None, metadata=field_options(alias="aantal_wielen")
    )
    odometer_judgement: VehicleOdometerJudgement | None = field(
        default=None,
        metadata=field_options(alias="tellerstandoordeel"),
    )
    pending_recall: bool | None = field(
        default=None,
        metadata=field_options(alias="openstaande_terugroepactie_indicator"),
    )
    taxi: bool | None = field(
        default=None, metadata=field_options(alias="taxi_indicator")
    )
    vehicle_type: VehicleType | None = field(
        default=None, metadata=field_options(alias="voertuigsoort")
    )

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Handle some fields that have some weirdness.

        Args:
        ----
            data: The values of the model.

        Returns:
        -------
            The adjusted values of the model.
        """
        # Convert certain values to None.
        for key in ("inrichting", "tellerstandoordeel", "voertuigsoort"):
            if d.get(key) in {"Niet geregistreerd", "N.v.t."}:
                d[key] = None

        # Make Brand and Model pretty
        for key in ("merk", "handelsbenaming"):
            d[key] = d[key].strip().title()

        return d
